"""Time related feature: date, holiday, event"""

import logging
from typing import List, Optional

import pandas as pd

from .registry import FeatureRegistry

logger = logging.getLogger(__name__)


def add_datetime_feature():
    return
