"""Auto regression feature: lag and roll"""

import logging

from .registry import FeatureRegistry

logger = logging.getLogger(__name__)


def add_transform_feature():
    # like log1p, sqrt
    return


def add_lag_feature():
    return


def add_roll_feature():
    return
