"""tfts Generator"""


class GenerationMixin:
    """
    A class containing all functions for auto-regressive generation, to be used as a mixin.
    """

    def generate(self, inputs, generation_config):
        return
